@@ -3,7 +3,7 @@
 
 
 # BleachBit
-# Copyright (C) 2008-2018 Andrew Ziem
+# Copyright (C) 2008-2020 Andrew Ziem
 # https://www.bleachbit.org
 #
 # This program is free software: you can redistribute it and/or modify
@@ -24,17 +24,14 @@
 Preferences dialog
 """
 
-from __future__ import absolute_import, print_function
-
 from bleachbit import _, _p, online_update_notification_enabled
 from bleachbit.Options import options
 from bleachbit import GuiBasic
+import bleachbit.GUI
 
-import gtk
+from gi.repository import Gtk, Gdk
 import logging
 import os
-import sys
-import traceback
 
 if 'nt' == os.name:
     from bleachbit import Windows
@@ -46,35 +43,52 @@ logger = logging.getLogger(__name__)
 LOCATIONS_WHITELIST = 1
 LOCATIONS_CUSTOM = 2
 
+win10_provider = None
+
 
 class PreferencesDialog:
 
     """Present the preferences dialog and save changes"""
 
-    def __init__(self, parent, cb_refresh_operations):
+    def __init__(self, parent, cb_refresh_operations, cb_set_windows10_theme):
         self.cb_refresh_operations = cb_refresh_operations
+        self.cb_set_windows10_theme = cb_set_windows10_theme
 
         self.parent = parent
-        self.dialog = gtk.Dialog(title=_("Preferences"),
-                                 parent=parent,
-                                 flags=gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT)
+        self.dialog = Gtk.Dialog(title=_("Preferences"),
+                                 transient_for=parent,
+                                 modal=True,
+                                 destroy_with_parent=True)
         self.dialog.set_default_size(300, 200)
 
-        notebook = gtk.Notebook()
-        notebook.append_page(self.__general_page(), gtk.Label(_("General")))
+        notebook = Gtk.Notebook()
+        notebook.append_page(self.__general_page(),
+                             Gtk.Label(label=_("General")))
         notebook.append_page(self.__locations_page(
-            LOCATIONS_CUSTOM), gtk.Label(_("Custom")))
-        notebook.append_page(self.__drives_page(), gtk.Label(_("Drives")))
+            LOCATIONS_CUSTOM), Gtk.Label(label=_("Custom")))
+        notebook.append_page(self.__drives_page(),
+                             Gtk.Label(label=_("Drives")))
         if 'posix' == os.name:
-            notebook.append_page(
-                self.__languages_page(), gtk.Label(_("Languages")))
+            notebook.append_page(self.__languages_page(),
+                                 Gtk.Label(label=_("Languages")))
         notebook.append_page(self.__locations_page(
-            LOCATIONS_WHITELIST), gtk.Label(_("Whitelist")))
+            LOCATIONS_WHITELIST), Gtk.Label(label=_("Whitelist")))
 
-        self.dialog.vbox.pack_start(notebook, True)
-        self.dialog.add_button(gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE)
+        # pack_start parameters: child, expand (reserve space), fill (actually fill it), padding
+        self.dialog.get_content_area().pack_start(notebook, True, True, 0)
+        self.dialog.add_button(Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE)
+
+        self.refresh_operations = False
+
+    def __del__(self):
+        """Destructor called when the dialog is closing"""
+        if self.refresh_operations:
+            # refresh the list of cleaners
+            self.cb_refresh_operations()
 
     def __toggle_callback(self, cell, path):
+        global win10_provider
+
         """Callback function to toggle option"""
         options.toggle(path)
         if online_update_notification_enabled:
@@ -83,110 +97,119 @@ class PreferencesDialog:
                 self.cb_winapp2.set_sensitive(
                     options.get('check_online_updates'))
         if 'auto_hide' == path:
-            self.cb_refresh_operations()
-        if 'auto_start' == path:
-            if 'nt' == os.name:
-                swc = Windows.start_with_computer
-            if 'posix' == os.name:
-                swc = Unix.start_with_computer
-            try:
-                swc(options.get(path))
-            except:
-                traceback.print_exc()
-                dlg = gtk.MessageDialog(self.parent,
-                                        type=gtk.MESSAGE_ERROR,
-                                        buttons=gtk.BUTTONS_OK,
-                                        message_format=str(sys.exc_info()[1]))
-                dlg.run()
-                dlg.destroy()
+            self.refresh_operations = True
+        if 'dark_mode' == path:
+            if os.name != 'nt' or options.get("win10_theme") == False:
+                Gtk.Settings.get_default().set_property(
+                    'gtk-application-prefer-dark-theme', options.get('dark_mode'))
+        if 'win10_theme' == path:
+            self.cb_set_windows10_theme()
+            self.cb_dark_mode.set_sensitive(not options.get('win10_theme'))
+        if 'debug' == path:
+            from bleachbit.Log import set_root_log_level
+            set_root_log_level()
 
     def __general_page(self):
         """Return a widget containing the general page"""
 
-        if 'nt' == os.name:
-            swcc = Windows.start_with_computer_check
-        if 'posix' == os.name:
-            swcc = Unix.start_with_computer_check
-
-        options.set('auto_start', swcc())
-
-        vbox = gtk.VBox()
+        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
 
         if online_update_notification_enabled:
-            cb_updates = gtk.CheckButton(
+            cb_updates = Gtk.CheckButton.new_with_label(
                 _("Check periodically for software updates via the Internet"))
             cb_updates.set_active(options.get('check_online_updates'))
             cb_updates.connect(
                 'toggled', self.__toggle_callback, 'check_online_updates')
             cb_updates.set_tooltip_text(
                 _("If an update is found, you will be given the option to view information about it.  Then, you may manually download and install the update."))
-            vbox.pack_start(cb_updates, False)
+            vbox.pack_start(cb_updates, False, True, 0)
 
-            updates_box = gtk.VBox()
+            updates_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
             updates_box.set_border_width(10)
 
-            self.cb_beta = gtk.CheckButton(_("Check for new beta releases"))
+            self.cb_beta = Gtk.CheckButton.new_with_label(
+                label=_("Check for new beta releases"))
             self.cb_beta.set_active(options.get('check_beta'))
             self.cb_beta.set_sensitive(options.get('check_online_updates'))
             self.cb_beta.connect(
                 'toggled', self.__toggle_callback, 'check_beta')
-            updates_box.pack_start(self.cb_beta, False)
+            updates_box.pack_start(self.cb_beta, False, True, 0)
 
             if 'nt' == os.name:
-                self.cb_winapp2 = gtk.CheckButton(
+                self.cb_winapp2 = Gtk.CheckButton.new_with_label(
                     _("Download and update cleaners from community (winapp2.ini)"))
                 self.cb_winapp2.set_active(options.get('update_winapp2'))
                 self.cb_winapp2.set_sensitive(
                     options.get('check_online_updates'))
                 self.cb_winapp2.connect(
                     'toggled', self.__toggle_callback, 'update_winapp2')
-                updates_box.pack_start(self.cb_winapp2, False)
-
-            vbox.pack_start(updates_box, False)
+                updates_box.pack_start(self.cb_winapp2, False, True, 0)
+            vbox.pack_start(updates_box, False, True, 0)
 
         # TRANSLATORS: This means to hide cleaners which would do
         # nothing.  For example, if Abrowser were never used on
         # this system, this option would hide Abrowser to simplify
         # the list of cleaners.
-        cb_auto_hide = gtk.CheckButton(_("Hide irrelevant cleaners"))
+        cb_auto_hide = Gtk.CheckButton.new_with_label(
+            label=_("Hide irrelevant cleaners"))
         cb_auto_hide.set_active(options.get('auto_hide'))
         cb_auto_hide.connect('toggled', self.__toggle_callback, 'auto_hide')
-        vbox.pack_start(cb_auto_hide, False)
+        vbox.pack_start(cb_auto_hide, False, True, 0)
 
         # TRANSLATORS: Overwriting is the same as shredding.  It is a way
         # to prevent recovery of the data. You could also translate
         # 'Shred files to prevent recovery.'
-        cb_shred = gtk.CheckButton(_("Overwrite contents of files to prevent recovery"))
+        cb_shred = Gtk.CheckButton(
+            label=_("Overwrite contents of files to prevent recovery"))
         cb_shred.set_active(options.get('shred'))
         cb_shred.connect('toggled', self.__toggle_callback, 'shred')
         cb_shred.set_tooltip_text(
             _("Overwriting is ineffective on some file systems and with certain BleachBit operations.  Overwriting is significantly slower."))
-        vbox.pack_start(cb_shred, False)
-
-        cb_start = gtk.CheckButton(_("Start BleachBit with computer"))
-        cb_start.set_active(options.get('auto_start'))
-        cb_start.connect('toggled', self.__toggle_callback, 'auto_start')
-        vbox.pack_start(cb_start, False)
+        vbox.pack_start(cb_shred, False, True, 0)
 
         # Close the application after cleaning is complete.
-        cb_exit = gtk.CheckButton(_("Exit after cleaning"))
+        cb_exit = Gtk.CheckButton.new_with_label(
+            label=_("Exit after cleaning"))
         cb_exit.set_active(options.get('exit_done'))
         cb_exit.connect('toggled', self.__toggle_callback, 'exit_done')
-        vbox.pack_start(cb_exit, False)
+        vbox.pack_start(cb_exit, False, True, 0)
 
         # Disable delete confirmation message.
-        cb_popup = gtk.CheckButton(_("Confirm before delete"))
+        cb_popup = Gtk.CheckButton(label=_("Confirm before delete"))
         cb_popup.set_active(options.get('delete_confirmation'))
         cb_popup.connect(
             'toggled', self.__toggle_callback, 'delete_confirmation')
-        vbox.pack_start(cb_popup, False)
+        vbox.pack_start(cb_popup, False, True, 0)
 
         # Use base 1000 over 1024?
-        cb_units_iec = gtk.CheckButton(
-            _("Use IEC sizes (1 KiB = 1024 bytes) instead of SI (1 kB = 1000 bytes)"))
+        cb_units_iec = Gtk.CheckButton(
+            label=_("Use IEC sizes (1 KiB = 1024 bytes) instead of SI (1 kB = 1000 bytes)"))
         cb_units_iec.set_active(options.get("units_iec"))
         cb_units_iec.connect('toggled', self.__toggle_callback, 'units_iec')
-        vbox.pack_start(cb_units_iec, False)
+        vbox.pack_start(cb_units_iec, False, True, 0)
+
+        if 'nt' == os.name:
+            # Dark theme
+            cb_win10_theme = Gtk.CheckButton(_("Windows 10 theme"))
+            cb_win10_theme.set_active(options.get("win10_theme"))
+            cb_win10_theme.connect(
+                'toggled', self.__toggle_callback, 'win10_theme')
+            vbox.pack_start(cb_win10_theme, False, True, 0)
+
+        # Dark theme
+        self.cb_dark_mode = Gtk.CheckButton(label=_("Dark mode"))
+        self.cb_dark_mode.set_active(options.get("dark_mode"))
+        self.cb_dark_mode.connect('toggled', self.__toggle_callback, 'dark_mode')
+        if 'nt' == os.name:
+            self.cb_dark_mode.set_sensitive(not options.get('win10_theme'))
+        vbox.pack_start(self.cb_dark_mode, False, True, 0)
+
+        # Debug logging
+        cb_debug = Gtk.CheckButton(label=_("Show debug messages"))
+        cb_debug.set_active(options.get("debug"))
+        cb_debug.connect('toggled', self.__toggle_callback, 'debug')
+        vbox.pack_start(cb_debug, False, True, 0)
+
         return vbox
 
     def __drives_page(self):
@@ -195,8 +218,8 @@ class PreferencesDialog:
         def add_drive_cb(button):
             """Callback for adding a drive"""
             title = _("Choose a folder")
-            pathname = GuiBasic.browse_folder(self.parent, title,
-                                              multiple=False, stock_button=gtk.STOCK_ADD)
+            pathname = GuiBasic.browse_folder(
+                self.parent, title, multiple=False, stock_button=Gtk.STOCK_ADD)
             if pathname:
                 liststore.append([pathname])
                 pathnames.append(pathname)
@@ -214,15 +237,15 @@ class PreferencesDialog:
             pathnames.remove(pathname)
             options.set_list('shred_drives', pathnames)
 
-        vbox = gtk.VBox()
+        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
 
         # TRANSLATORS: 'free' means 'unallocated'
-        notice = gtk.Label(
-            _("Choose a writable folder for each drive for which to overwrite free space."))
+        notice = Gtk.Label(label=_(
+            "Choose a writable folder for each drive for which to overwrite free space."))
         notice.set_line_wrap(True)
-        vbox.pack_start(notice, False)
+        vbox.pack_start(notice, False, True, 0)
 
-        liststore = gtk.ListStore(str)
+        liststore = Gtk.ListStore(str)
 
         pathnames = options.get_list('shred_drives')
         if pathnames:
@@ -231,27 +254,27 @@ class PreferencesDialog:
             pathnames = []
         for pathname in pathnames:
             liststore.append([pathname])
-        treeview = gtk.TreeView(model=liststore)
-        crt = gtk.CellRendererText()
-        tvc = gtk.TreeViewColumn(None, crt, text=0)
+        treeview = Gtk.TreeView.new_with_model(liststore)
+        crt = Gtk.CellRendererText()
+        tvc = Gtk.TreeViewColumn(None, crt, text=0)
         treeview.append_column(tvc)
 
-        vbox.pack_start(treeview)
+        vbox.pack_start(treeview, True, True, 0)
 
         # TRANSLATORS: In the preferences dialog, this button adds a path to
         # the list of paths
-        button_add = gtk.Button(_p('button', 'Add'))
+        button_add = Gtk.Button.new_with_label(label=_p('button', 'Add'))
         button_add.connect("clicked", add_drive_cb)
         # TRANSLATORS: In the preferences dialog, this button removes a path
         # from the list of paths
-        button_remove = gtk.Button(_p('button', 'Remove'))
+        button_remove = Gtk.Button.new_with_label(label=_p('button', 'Remove'))
         button_remove.connect("clicked", remove_drive_cb)
 
-        button_box = gtk.HButtonBox()
-        button_box.set_layout(gtk.BUTTONBOX_START)
-        button_box.pack_start(button_add)
-        button_box.pack_start(button_remove)
-        vbox.pack_start(button_box, False)
+        button_box = Gtk.ButtonBox(orientation=Gtk.Orientation.HORIZONTAL)
+        button_box.set_layout(Gtk.ButtonBoxStyle.START)
+        button_box.pack_start(button_add, True, True, 0)
+        button_box.pack_start(button_remove, True, True, 0)
+        vbox.pack_start(button_box, False, True, 0)
 
         return vbox
 
@@ -266,43 +289,42 @@ class PreferencesDialog:
             langid = liststore[path][1]
             options.set_language(langid, value)
 
-        vbox = gtk.VBox()
-
-        notice = gtk.Label(
-            _("All languages will be deleted except those checked."))
-        vbox.pack_start(notice, False)
+        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
+        notice = Gtk.Label(
+            label=_("All languages will be deleted except those checked."))
+        vbox.pack_start(notice, False, False, 0)
 
         # populate data
-        liststore = gtk.ListStore('gboolean', str, str)
+        liststore = Gtk.ListStore('gboolean', str, str)
         for lang, native in sorted(Unix.Locales.native_locale_names.items()):
             liststore.append([(options.get_language(lang)), lang, native])
 
         # create treeview
-        treeview = gtk.TreeView(liststore)
+        treeview = Gtk.TreeView.new_with_model(liststore)
 
         # create column views
-        self.renderer0 = gtk.CellRendererToggle()
+        self.renderer0 = Gtk.CellRendererToggle()
         self.renderer0.set_property('activatable', True)
         self.renderer0.connect('toggled', preserve_toggled_cb, liststore)
-        self.column0 = gtk.TreeViewColumn(
+        self.column0 = Gtk.TreeViewColumn(
             _("Preserve"), self.renderer0, active=0)
         treeview.append_column(self.column0)
 
-        self.renderer1 = gtk.CellRendererText()
-        self.column1 = gtk.TreeViewColumn(_("Code"), self.renderer1, text=1)
+        self.renderer1 = Gtk.CellRendererText()
+        self.column1 = Gtk.TreeViewColumn(_("Code"), self.renderer1, text=1)
         treeview.append_column(self.column1)
 
-        self.renderer2 = gtk.CellRendererText()
-        self.column2 = gtk.TreeViewColumn(_("Name"), self.renderer2, text=2)
+        self.renderer2 = Gtk.CellRendererText()
+        self.column2 = Gtk.TreeViewColumn(_("Name"), self.renderer2, text=2)
         treeview.append_column(self.column2)
         treeview.set_search_column(2)
 
         # finish
-        swindow = gtk.ScrolledWindow()
-        swindow.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
+        swindow = Gtk.ScrolledWindow()
+        swindow.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
         swindow.set_size_request(300, 200)
         swindow.add(treeview)
-        vbox.pack_start(swindow)
+        vbox.pack_start(swindow, False, True, 0)
         return vbox
 
     def __locations_page(self, page_type):
@@ -315,7 +337,8 @@ class PreferencesDialog:
             if pathname:
                 for this_pathname in pathnames:
                     if pathname == this_pathname[1]:
-                        logger.warning("'%s' already exists in whitelist", pathname)
+                        logger.warning(
+                            "'%s' already exists in whitelist", pathname)
                         return
                 liststore.append([_('File'), pathname])
                 pathnames.append(['file', pathname])
@@ -325,11 +348,12 @@ class PreferencesDialog:
             """Callback for adding a folder"""
             title = _("Choose a folder")
             pathname = GuiBasic.browse_folder(self.parent, title,
-                                              multiple=False, stock_button=gtk.STOCK_ADD)
+                                              multiple=False, stock_button=Gtk.STOCK_ADD)
             if pathname:
                 for this_pathname in pathnames:
                     if pathname == this_pathname[1]:
-                        logger.warning("'%s' already exists in whitelist", pathname)
+                        logger.warning(
+                            "'%s' already exists in whitelist", pathname)
                         return
                 liststore.append([_('Folder'), pathname])
                 pathnames.append(['folder', pathname])
@@ -356,7 +380,8 @@ class PreferencesDialog:
             if pathname:
                 for this_pathname in pathnames:
                     if pathname == this_pathname[1]:
-                        logger.warning("'%s' already exists in whitelist", pathname)
+                        logger.warning(
+                            "'%s' already exists in whitelist", pathname)
                         return
                 liststore.append([_('File'), pathname])
                 pathnames.append(['file', pathname])
@@ -366,11 +391,12 @@ class PreferencesDialog:
             """Callback for adding a folder"""
             title = _("Choose a folder")
             pathname = GuiBasic.browse_folder(self.parent, title,
-                                              multiple=False, stock_button=gtk.STOCK_ADD)
+                                              multiple=False, stock_button=Gtk.STOCK_ADD)
             if pathname:
                 for this_pathname in pathnames:
                     if pathname == this_pathname[1]:
-                        logger.warning("'%s' already exists in whitelist", pathname)
+                        logger.warning(
+                            "'%s' already exists in whitelist", pathname)
                         return
                 liststore.append([_('Folder'), pathname])
                 pathnames.append(['folder', pathname])
@@ -390,14 +416,14 @@ class PreferencesDialog:
                     pathnames.remove(this_pathname)
                     options.set_custom_paths(pathnames)
 
-        vbox = gtk.VBox()
+        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
 
         # load data
         if LOCATIONS_WHITELIST == page_type:
             pathnames = options.get_whitelist_paths()
         elif LOCATIONS_CUSTOM == page_type:
             pathnames = options.get_custom_paths()
-        liststore = gtk.ListStore(str, str)
+        liststore = Gtk.ListStore(str, str)
         for paths in pathnames:
             type_code = paths[0]
             type_str = None
@@ -413,60 +439,63 @@ class PreferencesDialog:
         if LOCATIONS_WHITELIST == page_type:
             # TRANSLATORS: "Paths" is used generically to refer to both files
             # and folders
-            notice = gtk.Label(
-                _("Theses paths will not be deleted or modified."))
+            notice = Gtk.Label(
+                label=_("Theses paths will not be deleted or modified."))
         elif LOCATIONS_CUSTOM == page_type:
-            notice = gtk.Label(
-                _("These locations can be selected for deletion."))
-        vbox.pack_start(notice, False)
+            notice = Gtk.Label(
+                label=_("These locations can be selected for deletion."))
+        vbox.pack_start(notice, False, False, 0)
 
         # create treeview
-        treeview = gtk.TreeView(liststore)
+        treeview = Gtk.TreeView.new_with_model(liststore)
 
         # create column views
-        self.renderer0 = gtk.CellRendererText()
-        self.column0 = gtk.TreeViewColumn(_("Type"), self.renderer0, text=0)
+        self.renderer0 = Gtk.CellRendererText()
+        self.column0 = Gtk.TreeViewColumn(_("Type"), self.renderer0, text=0)
         treeview.append_column(self.column0)
 
-        self.renderer1 = gtk.CellRendererText()
+        self.renderer1 = Gtk.CellRendererText()
         # TRANSLATORS: In the tree view "Path" is used generically to refer to a
         # file, a folder, or a pattern describing either
-        self.column1 = gtk.TreeViewColumn(_("Path"), self.renderer1, text=1)
+        self.column1 = Gtk.TreeViewColumn(_("Path"), self.renderer1, text=1)
         treeview.append_column(self.column1)
         treeview.set_search_column(1)
 
         # finish tree view
-        swindow = gtk.ScrolledWindow()
-        swindow.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
+        swindow = Gtk.ScrolledWindow()
+        swindow.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
         swindow.set_size_request(300, 200)
         swindow.add(treeview)
-        vbox.pack_start(swindow)
+
+        vbox.pack_start(swindow, False, True, 0)
 
         # buttons that modify the list
-        button_add_file = gtk.Button(_p('button', 'Add file'))
+        button_add_file = Gtk.Button.new_with_label(
+            label=_p('button', 'Add file'))
         if LOCATIONS_WHITELIST == page_type:
             button_add_file.connect("clicked", add_whitelist_file_cb)
         elif LOCATIONS_CUSTOM == page_type:
             button_add_file.connect("clicked", add_custom_file_cb)
 
-        button_add_folder = gtk.Button(_p('button', 'Add folder'))
+        button_add_folder = Gtk.Button.new_with_label(
+            label=_p('button', 'Add folder'))
         if LOCATIONS_WHITELIST == page_type:
             button_add_folder.connect("clicked", add_whitelist_folder_cb)
         elif LOCATIONS_CUSTOM == page_type:
             button_add_folder.connect("clicked", add_custom_folder_cb)
 
-        button_remove = gtk.Button(_p('button', 'Remove'))
+        button_remove = Gtk.Button.new_with_label(label=_p('button', 'Remove'))
         if LOCATIONS_WHITELIST == page_type:
             button_remove.connect("clicked", remove_whitelist_path_cb)
         elif LOCATIONS_CUSTOM == page_type:
             button_remove.connect("clicked", remove_custom_path_cb)
 
-        button_box = gtk.HButtonBox()
-        button_box.set_layout(gtk.BUTTONBOX_START)
-        button_box.pack_start(button_add_file)
-        button_box.pack_start(button_add_folder)
-        button_box.pack_start(button_remove)
-        vbox.pack_start(button_box, False)
+        button_box = Gtk.ButtonBox(orientation=Gtk.Orientation.HORIZONTAL)
+        button_box.set_layout(Gtk.ButtonBoxStyle.START)
+        button_box.pack_start(button_add_file, True, True, 0)
+        button_box.pack_start(button_add_folder, True, True, 0)
+        button_box.pack_start(button_remove, True, True, 0)
+        vbox.pack_start(button_box, False, True, 0)
 
         # return page
         return vbox
